@@ -26,7 +26,9 @@
 #include <gtk/gtk.h>
 
 #include <libgames-support/games-card-images.h>
+#include <libgames-support/games-files.h>
 #include <libgames-support/games-pixbuf-utils.h>
+#include <libgames-support/games-sound.h>
 
 #include "conf.h"
 
@@ -170,6 +172,7 @@ struct _AisleriotBoardPrivate
   guint click_status : 4; /* enough bits for MoveStatus */
 
   guint show_selection : 1;
+  guint show_highlight : 1;
   guint scalable_cards : 1;
 };
 
@@ -342,7 +345,7 @@ set_background_from_baize (GtkWidget *widget,
   char *path;
   int width, height;
 
-  path = g_build_filename (PIXMAPDIR, "baize.png", NULL);
+  path = games_build_filename (PIXMAPDIR, "baize.png");
 
   pixbuf = gdk_pixbuf_new_from_file (path, &error);
   g_free (path);
@@ -828,7 +831,8 @@ slot_update_card_images (AisleriotBoard *board,
   AisleriotBoardPrivate *priv = board->priv;
   int highlight_start_card_id = G_MAXINT;
 
-  if (G_UNLIKELY (slot == priv->highlight_slot)) {
+  if (G_UNLIKELY (slot == priv->highlight_slot &&
+                  priv->show_highlight)) {
     highlight_start_card_id = slot->cards->len - 1;
   } else if (G_UNLIKELY (slot == priv->selection_slot &&
                          priv->selection_start_card_id >= 0 &&
@@ -933,11 +937,16 @@ drag_begin (AisleriotBoard *board)
   GdkDrawable *drawable;
   GdkWindowAttr attributes;
 
+  if (!priv->selection_slot ||
+      priv->selection_start_card_id < 0) {
+    priv->click_status = STATUS_NONE;
+    return;
+  }
+
   priv->click_status = STATUS_IS_DRAG;
 
   hslot = priv->moving_cards_origin_slot = priv->selection_slot;
   priv->moving_cards_origin_card_id = priv->selection_start_card_id;
-  g_assert (hslot != NULL && priv->moving_cards_origin_card_id >= 0);
 
   num_moving_cards = hslot->cards->len - priv->moving_cards_origin_card_id;
 
@@ -1072,6 +1081,8 @@ drag_begin (AisleriotBoard *board)
 
   slot_update_geometry (board, hslot);
   slot_update_card_images (board, hslot);
+      
+  set_cursor (board, CURSOR_CLOSED);
 }
 
 static void
@@ -1104,14 +1115,17 @@ drag_end (AisleriotBoard *board,
 
 static gboolean
 cards_are_droppable (AisleriotBoard *board,
-                     Slot *hslot)
+                     Slot *slot)
 {
   AisleriotBoardPrivate *priv = board->priv;
 
-  return hslot && priv->moving_cards_origin_slot &&
-    aisleriot_game_drop_valid (priv->game, priv->moving_cards_origin_slot->id, hslot->id,
-                               priv->moving_cards->data,
-                               priv->moving_cards->len);
+  return slot != NULL &&
+         priv->moving_cards_origin_slot &&
+         aisleriot_game_drop_valid (priv->game,
+                                    priv->moving_cards_origin_slot->id,
+                                    slot->id,
+                                    priv->moving_cards->data,
+                                    priv->moving_cards->len);
 }
 
 static Slot *
@@ -1131,7 +1145,7 @@ find_drop_target (AisleriotBoard *board,
                                 y + priv->card_size.height / 2,
                                 &new_hslot, &new_cardid);
 
-  if (new_hslot && cards_are_droppable (board, new_hslot))
+  if (cards_are_droppable (board, new_hslot))
     return new_hslot;
 
   /* If that didn't work, look for a target at all 4 corners of the card. */
@@ -1186,8 +1200,10 @@ drop_moving_cards (AisleriotBoard *board,
 
   if (moved) {
     aisleriot_game_end_move (priv->game);
+    games_sound_play ("click");
   } else {
     aisleriot_game_discard_move (priv->game);
+    games_sound_play ("slide");
   }
 
   drag_end (board, moved);
@@ -1208,7 +1224,8 @@ highlight_drop_target (AisleriotBoard *board,
     return;
 
   /* Invalidate the old highlight rect */
-  if (priv->highlight_slot != NULL) {
+  if (priv->highlight_slot != NULL &&
+      priv->show_highlight) {
     get_rect_by_slot_and_card (board,
                                priv->highlight_slot,
                                priv->highlight_slot->cards->len - 1 /* it's ok if this is == -1 */,
@@ -1221,10 +1238,16 @@ highlight_drop_target (AisleriotBoard *board,
      */
     slot_update_card_images_full (board, priv->highlight_slot, G_MAXINT);
   }
-    
+
+  /* Need to set the highlight slot even when we the cards aren't droppable
+   * since that can happen when the game doesn't support FEATURE_DROPPABLE.
+   */
   priv->highlight_slot = slot;
   
-  if (slot == NULL || !cards_are_droppable (board, slot))
+  if (!cards_are_droppable (board, slot))
+    return;
+
+  if (!priv->show_highlight)
     return;
 
   /* Prepare the highlight pixbuf/pixmaps and invalidate the new highlight rect */
@@ -1301,16 +1324,16 @@ clear_state (AisleriotBoard *board)
 /* Game state handling */
 
 static void
-sync_game_features (AisleriotGame *game,
-                    GParamSpec *pspec,
-                    AisleriotBoard *board)
+game_type_changed_cb (AisleriotGame *game,
+                      AisleriotBoard *board)
 {
   AisleriotBoardPrivate *priv = board->priv;
-  guint features = 0;
+  guint features;
 
-  g_object_get (game, "features", &features, NULL);
+  features = aisleriot_game_get_features (game);
 
   priv->droppable_supported = ((features & FEATURE_DROPPABLE) != 0);
+  priv->show_highlight = priv->droppable_supported;
 }
 
 static void
@@ -1365,8 +1388,22 @@ slot_changed_cb (AisleriotGame *game,
   slot_update_geometry (board, slot);
   slot_update_card_images (board, slot);
 
+  if (slot == priv->moving_cards_origin_slot) {
+    /* PANIC! */
+    /* FIXMEchpe */
+  }
   if (slot == priv->selection_slot) {
     set_selection (board, NULL, -1);
+
+    /* If this slot changes while we're in a click cycle, abort the action.
+     * That prevents a problem where the cards that were selected and
+     * about to be dragged have vanished because of autoplay; c.f. bug #449767.
+     * Note: we don't use clear_state() here since we only want to disable
+     * the highlight and revealed card if that particular slot changed, see
+     * the code above. And we don't clear last_clicked_slot/card_id either, so
+     * a double-click will still work.
+     */
+    priv->click_status = STATUS_NONE;
   }
   if (slot == priv->focus_slot) {
     set_focus (board, slot, priv->focus_card_id, priv->show_focus);
@@ -1389,8 +1426,10 @@ aisleriot_board_realize (GtkWidget *widget)
   GtkSettings *settings;
 #ifndef HAVE_MAEMO
   gboolean touchscreen_mode;
+#ifdef GDK_WINDOWING_X11
   char *xft_rgba = NULL;
-#endif
+#endif /* GDK_WINDOWING_X11 */
+#endif /* !HAVE_MAEMO */
 
   GTK_WIDGET_CLASS (aisleriot_board_parent_class)->realize (widget);
 
@@ -1420,11 +1459,13 @@ aisleriot_board_realize (GtkWidget *widget)
                 "gtk-double-click-time", &priv->double_click_time,
 #ifndef HAVE_MAEMO
                 "gtk-touchscreen-mode", &touchscreen_mode,
+#ifdef GDK_WINDOWING_X11
                 "gtk-xft-rgba", &xft_rgba,
-#endif
+#endif /* GDK_WINDOWING_X11 */
+#endif /* !HAVE_MAEMO */
                 NULL);
 
-#ifndef HAVE_MAEMO
+#if defined (GDK_WINDOWING_X11) && !defined (HAVE_MAEMO)
   if (xft_rgba != NULL) {
     gboolean antialias_set = TRUE;
     cairo_antialias_t antialias_mode = CAIRO_ANTIALIAS_DEFAULT;
@@ -1454,9 +1495,10 @@ aisleriot_board_realize (GtkWidget *widget)
                                        subpixel_order);
     }
   }
-
+#endif /* GDK_WINDOWING_X11 && !HAVE_MAEMO */
+#ifndef HAVE_MAEMO
   priv->touchscreen_mode = touchscreen_mode != FALSE;
-#endif
+#endif /* !HAVE_MAEMO */
 
   aisleriot_board_setup_geometry (board);
 }
@@ -1553,6 +1595,13 @@ static gboolean
 aisleriot_board_focus_in (GtkWidget *widget,
                           GdkEventFocus *event)
 {
+  AisleriotBoard *board = AISLERIOT_BOARD (widget);
+  AisleriotBoardPrivate *priv = board->priv;
+
+  /* If we're not showing focus, no need to redraw on focus change */
+  if (!priv->show_focus)
+    return FALSE;
+
   return GTK_WIDGET_CLASS (aisleriot_board_parent_class)->focus_in_event (widget, event);
 }
 
@@ -1561,9 +1610,14 @@ aisleriot_board_focus_out (GtkWidget *widget,
                            GdkEventFocus *event)
 {
   AisleriotBoard *board = AISLERIOT_BOARD (widget);
+  AisleriotBoardPrivate *priv = board->priv;
 
   clear_state (board);
 
+  /* If we're not showing focus, no need to redraw on focus change */
+  if (!priv->show_focus)
+    return FALSE;
+
   return GTK_WIDGET_CLASS (aisleriot_board_parent_class)->focus_out_event (widget, event);
 }
 
@@ -1685,19 +1739,20 @@ aisleriot_board_button_press (GtkWidget *widget,
     guint8 *cards;
     guint n_cards;
 
-    /* Check if we should move */
+    /* NOTE: We cannot use aisleriot_game_drop_valid here since the
+     * game may not support the "droppable" feature.
+     */
     if (!priv->click_to_move ||
-        selection_start_card_id < 0 ||
-        !aisleriot_game_drop_valid (priv->game,
-                                    selection_slot->id,
-                                    hslot->id,
-                                    selection_slot->cards->data + selection_start_card_id,
-                                    selection_slot->cards->len - selection_start_card_id))
+        selection_start_card_id < 0)
       goto set_selection;
 
-    /* Unset the selection; we'll re-set it after the move */
+    /* Remove the old selection. If the move doesn't succeed, we'll select
+     * the clicked-on cards instead.
+     */
     set_selection (board, NULL, -1);
 
+    priv->click_status = STATUS_NONE;
+
     aisleriot_game_record_move (priv->game,
                                 selection_slot->id,
                                 selection_slot->cards->data,
@@ -1728,19 +1783,18 @@ aisleriot_board_button_press (GtkWidget *widget,
         g_signal_emit_by_name (priv->game, "slot-changed", selection_slot); /* FIXMEchpe! */
 
       aisleriot_game_test_end_of_game (priv->game);
-    } else {
-      aisleriot_game_discard_move (priv->game);
 
-      /* FIXMEchpe: maybe beep? */
-
-      /* Add the cards back to the origin slot, and select them */
-      aisleriot_game_slot_add_cards (priv->game, selection_slot, cards, n_cards);
-      set_selection_with_cards (board, selection_slot, cards, n_cards);
+      return TRUE;
     }
 
-    priv->click_status = STATUS_NONE;
+    /* Not moved; discard the move and select the new cards */
+    aisleriot_game_discard_move (priv->game);
 
-    return TRUE;
+    aisleriot_game_slot_add_cards (priv->game, selection_slot, cards, n_cards);
+
+    /* FIXMEchpe: maybe beep? */
+
+    goto set_selection;
   }
 
   if (cardid != priv->selection_start_card_id)
@@ -1829,6 +1883,7 @@ aisleriot_board_button_release (GtkWidget *widget,
       aisleriot_game_record_move (priv->game, -1, NULL, 0);
       if (aisleriot_game_button_clicked_lambda (priv->game, slot->id)) {
         aisleriot_game_end_move (priv->game);
+	games_sound_play ("click");
       } else {
         aisleriot_game_discard_move (priv->game);
       }
@@ -1882,7 +1937,6 @@ aisleriot_board_motion_notify (GtkWidget *widget,
                                        event->x,
                                        event->y)) {
     drag_begin (board);
-    set_cursor (board, CURSOR_CLOSED);
   } else {
     set_cursor_by_location (board, event->x, event->y);
   }
@@ -2042,7 +2096,8 @@ aisleriot_board_expose_event (GtkWidget *widget,
       if (G_LIKELY (hslot != highlight_slot)) {
         pixbuf = priv->slot_image;
       } else {
-        pixbuf = games_card_images_get_slot_pixbuf (priv->images, TRUE);
+        pixbuf = games_card_images_get_slot_pixbuf (priv->images,
+                                                    priv->show_highlight);
       }
 
       if (!pixbuf)
@@ -2058,7 +2113,8 @@ aisleriot_board_expose_event (GtkWidget *widget,
       if (G_LIKELY (hslot != highlight_slot)) {
         pixmap = priv->slot_image;
       } else {
-        pixmap = games_card_images_get_slot_pixmap (priv->images, TRUE);
+        pixmap = games_card_images_get_slot_pixmap (priv->images,
+                                                    priv->show_highlight);
       }
 
       if (!pixmap)
@@ -2283,15 +2339,15 @@ aisleriot_board_set_property (GObject *object,
   switch (prop_id) {
     case PROP_GAME:
       priv->game = AISLERIOT_GAME (g_value_dup_object (value));
-      sync_game_features (priv->game, NULL, board);
+
+      g_signal_connect (priv->game, "game-type",
+                        G_CALLBACK (game_type_changed_cb), board);
       g_signal_connect (priv->game, "game-cleared",
                         G_CALLBACK (game_cleared_cb), board);
       g_signal_connect (priv->game, "game-new",
                         G_CALLBACK (game_new_cb), board);
       g_signal_connect (priv->game, "slot-changed",
                         G_CALLBACK (slot_changed_cb), board);
-      g_signal_connect (priv->game, "notify::features",
-                        G_CALLBACK (sync_game_features), board);
 
       break;
     case PROP_THEME:
@@ -2388,9 +2444,6 @@ aisleriot_board_set_card_theme (AisleriotBoard *board,
   priv->slot_image = NULL;
 
   retval = games_card_images_set_theme (priv->images, card_theme);
-  if (!retval) {
-    g_warning ("Loading theme %d failed!", card_theme);
-  }
 
   /* NOTE! We need to do this even if setting the theme failed, since
    * the attempt will have wiped out the old theme data!
